import time, re, os, sys
from datetime import datetime
from lib import IPhoenix
from TMDB import TMDB
from thetvdbapi import TheTVDB
import threading
try:
    import xbmc, xbmcaddon
except:
    from dev.xbmclib import *
    
try:
    from sqlite3 import dbapi2 as sqlite
    # print "Loading sqlite3 as DB engine"
except:
    from pysqlite2 import dbapi2 as sqlite
    # print "Loading sqlite3 as DB engine"


class MetaData(IPhoenix):
    '''
    This class performs all the handling of meta data, requesting, storing and sending back to calling application
    
        - Create cache DB if it does not exist
        - Create a meta data zip file container to share
        - Get the meta data from TMDB/IMDB/TVDB
        - Store/Retrieve meta from cache DB
        - Download image files locally
    '''
    dbname = "MetaData.db"
    dbpath = None
    path = None
    cachePath = None
    tvpath = None
    tvcovers = None
    tvbackdrops = None
    tvbanners = None
    mvpath = None
    mvcovers = None
    mvbackdrops = None
    typeMovie = 'movie'
    typeTVShow = 'tvshow'
    typeSeason = 'season'
    typeEpisode = 'episode'
    
    def __init__(self, path='special://database/phoenix/core/', preparezip=False):
        
        self.dbpath = path
        self.path = xbmc.translatePath(path)
        #self.cache_path = self.__makeDir(self.path, 'meta_cache')
            
        # control whether class is being used to prepare pre-packaged .zip
        self.classmode = str((preparezip==True)).lower()
        #self.tvpath = self.__makeDir(self.cache_path, self.typeTVShow)
        #self.tvcovers = self.__makeDir(self.tvpath, 'covers')
        #self.tvbackdrops = self.__makeDir(self.tvpath, 'backdrops')
        #self.tvbanners = self.__makeDir(self.tvpath, 'banners')
        #self.mvpath = self.__makeDir(self.cache_path, self.typeMovie)
        #self.mvcovers = self.__makeDir(self.mvpath, 'covers')
        #self.mvbackdrops = self.__makeDir(self.mvpath, 'backdrops')
        
        IPhoenix.__init__(self)
        self.dbcur= self.getDBCursor()
        
        
        
    def __makeDir(self,mypath, dirname):        
        ''' Creates sub-directories if they are not found. '''
        subpath = os.path.join(mypath, dirname)
        if not os.path.exists(subpath): os.makedirs(subpath)
        return subpath
    
    def initDatabase(self, cursor):
        IPhoenix.initDatabase(self, cursor)
        self.getDBConnection().row_factory = sqlite.Row # return results indexed by field names and not numbers so we can convert to dict
        cursor.execute("""
            CREATE TABLE IF NOT EXISTS movie_meta (
                imdb_id VARCHAR(25), 
                tmdb_id VARCHAR(25), 
                title VARCHAR(512),
                year INTEGER,
                director TEXT,
                writer TEXT,
                tagline TEXT, 
                cast TEXT,
                rating FLOAT,
                votes TEXT,
                duration TEXT,
                plot TEXT,
                mpaa TEXT,
                premiered TEXT,
                genre TEXT,
                studio TEXT,
                thumb_url TEXT,
                cover_url TEXT,
                trailer_url TEXT,
                backdrop_url TEXT,
                imgs_prepacked VARCHAR(5),
                UNIQUE(imdb_id, tmdb_id, title, year)
            );""")
        cursor.execute("CREATE INDEX IF NOT EXISTS idx_movie_meta_title on movie_meta (title);")
        cursor.execute("CREATE INDEX IF NOT EXISTS idx_movie_meta_title_year on movie_meta (title,year);")
        cursor.execute("CREATE INDEX IF NOT EXISTS idx_movie_meta_imdb_id on movie_meta (imdb_id);")
        cursor.execute("CREATE INDEX IF NOT EXISTS idx_movie_meta_tmdb_id on movie_meta (tmdb_id);")
        cursor.execute(""" 
            CREATE TABLE IF NOT EXISTS tvshow_meta (
                imdb_id VARCHAR(25), 
                tvdb_id VARCHAR(25),
                title  VARCHAR(512), 
                cast TEXT,
                rating FLOAT, 
                duration TEXT, 
                plot TEXT,
                mpaa TEXT, 
                premiered TEXT, 
                genre TEXT, 
                studio TEXT,
                status TEXT,
                banner_url TEXT, 
                cover_url TEXT,
                trailer_url TEXT, 
                backdrop_url TEXT,
                imgs_prepacked VARCHAR(5),
                UNIQUE(imdb_id, tvdb_id, title)
            );""")
        cursor.execute("CREATE INDEX IF NOT EXISTS idx_tvshow_meta_title on tvshow_meta (title);")
        cursor.execute("CREATE INDEX IF NOT EXISTS idx_tvshow_meta_title on tvshow_meta (imdb_id);")
        cursor.execute("CREATE INDEX IF NOT EXISTS idx_tvshow_meta_title on tvshow_meta (tvdb_id);")
        cursor.execute("""  
            CREATE TABLE IF NOT EXISTS season_meta (
                imdb_id VARCHAR(25), 
                tvdb_id VARCHAR(25),
                season INTEGER,
                cover_url TEXT,
                UNIQUE(imdb_id, tvdb_id, season)
            );""")
        cursor.execute("CREATE INDEX IF NOT EXISTS idx_season_meta_imdb_id on season_meta (imdb_id);")
        cursor.execute("CREATE INDEX IF NOT EXISTS idx_season_meta_tvdb_id on season_meta (tvdb_id);")
        cursor.execute("CREATE INDEX IF NOT EXISTS idx_season_meta_season on season_meta (season);")
        cursor.execute("""
            CREATE TABLE IF NOT EXISTS episode_meta (
                imdb_id VARCHAR(25), 
                tvdb_id VARCHAR(25), 
                episode_id TEXT,
                season INTEGER,
                episode INTEGER, 
                title  VARCHAR(512), 
                director TEXT, 
                writer TEXT, 
                plot TEXT, 
                rating FLOAT, 
                premiered TEXT, 
                poster TEXT, 
                UNIQUE(imdb_id, tvdb_id, episode_id, title)
            );""")
        cursor.execute("CREATE INDEX IF NOT EXISTS idx_episode_meta_title on episode_meta (title);")
        cursor.execute("CREATE INDEX IF NOT EXISTS idx_episode_imdb_id_meta_season_episode on episode_meta (imdb_id,season,episode);")
        
        
    def zipDir(self,basedir, archivename):
        from contextlib import closing
        from zipfile import ZipFile, ZIP_DEFLATED
        assert os.path.isdir(basedir)
        with closing(ZipFile(archivename, "w", ZIP_DEFLATED)) as z:
            for root, dirs, files in os.walk(basedir):
                #NOTE: ignore empty directories
                for fn in files:
                    absfn = os.path.join(root, fn)
                    zfn = absfn[len(basedir)+len(os.sep):] #XXX: relative path
                    z.write(absfn, zfn)
                    
    def extractZip(self,src, dest):
        assert os.path.isfile(src)
        assert os.path.isdir(dest)
        try:
            #print 'Extracting '+str(src)+' to '+str(dest)
            #make sure there are no double slashes in paths
            src=os.path.normpath(src)
            dest=os.path.normpath(dest) 
            #Unzip - Only if file size is > 1KB
            if os.path.getsize(src) > 10000:
                xbmc.executebuiltin("XBMC.Extract("+src+","+dest+")")
            else:
                #print '************* Error: File size is too small'
                return False
        except:
            #print 'Extraction failed!'
            return False
        else:                
            #print 'Extraction success!'
            return True
        
    def createMetaPacks(self):
        pass
    
    def installMetaPack(self,pack):
        pass
    
    def __initTVShowMeta(self, imdb_id, tvdb_id, name):
        '''
        Initializes a tvshow_meta dictionary with default values, to ensure we always
        have all fields
        
        Args:
            imdb_id (str): IMDB ID
            name (str): full name of movie you are searching
            premiered (str): 10 digit year YYYY-MM-DD
                        
        Returns:
            DICT in the structure of what is required to write to the DB
        '''
        return {
                "imdb_id":str(imdb_id),
                "tvdb_id":str(tvdb_id),
                "title":str(name),
                "TVShowTitle":str(name),
                "rating":0,
                "duration":"",
                "plot":"",
                "mpaa":"",
                "premiered":"",
                "trailer_url":"",
                "genre":"",
                "studio":"",
                "status":"",
                "cast":[],
                "banner_url":"",
                "imgs_prepacked":self.classmode,
                "cover_url":"",
                "backdrop_url":"",
                "episode":"0"
            }
        
    def __initMovieMeta(self, imdb_id, tmdb_id, name, year=0):
        '''
        Initializes a movie_meta dictionary with default values, to ensure we always
        have all fields
        
        Args:
            imdb_id (str): IMDB ID
            name (str): full name of movie you are searching
            premiered (str): 10 digit year YYYY-MM-DD
                        
        Returns:
            DICT in the structure of what is required to write to the DB
        '''
        if year:
            year = int(year)
        else:
            year = 0
        return {
                "imdb_id":str(imdb_id),
                "tvdb_id":str(tmdb_id),
                "title":str(name),
                "year":year,
                "writer":"",
                "director":"",
                "tagline":"",
                "cast":[],
                "rating":0,
                "votes":"",
                "duration":"",
                "plot":"",
                "mpaa":"",
                "premiered":"",
                "trailer_url":"",
                "genre":"",
                "studio":"",
                "imgs_prepacked":self.classmode,
                "thumb_url":"",
                "cover_url":"",
                "backdrop_url":""
            }
        
    def __stringCompare(self, s1, s2):
        """ Method that takes two strings and returns True or False, based
            on if they are equal, regardless of case.
        """
        try:
            return s1.lower() == s2.lower()
        except AttributeError:
            #print "Please only pass strings into this method."
            #print "You passed a %s and %s" % (s1.__class__, s2.__class__)
            pass
   
    def __cleanString(self, string):
        """ 
            Method that takes a string and returns it cleaned of any special characters
            in order to do proper string comparisons
        """        
        try:
            return ''.join(e for e in string if e.isalnum())
        except:
            return string
 

    def __convertDate(self, string, in_format, out_format):
        ''' Helper method to convert a string date to a given format '''
        
        #Legacy check, Python 2.4 does not have strptime attribute, instroduced in 2.5
        if hasattr(datetime, 'strptime'):
            strptime = datetime.strptime
        else:
            strptime = lambda date_string, format: datetime(*(time.strptime(date_string, format)[0:6]))
        try:
            a = strptime(string, in_format).strftime(out_format)
        except Exception, e:
            #print '************* Error Date conversion failed: %s' % e
            return None
        return a
        
        
    def __getDate(self, year, month_day):
        month_name = month_day[:3]
        day=month_day[4:]
        month = {'Jan':'01','Feb':'01','Mar':'01','Apr':'01','May':'01','Jun':'01','Jul':'01','Aug':'01','Sep':'01','Oct':'01','Nov':'01','Dec':'01'}.get(month_name,"01")               
        return year + '-' + month + '-' + day
    
    
    def __downloadImage(self, url, path, name):
        '''
        Download images to save locally
        
        Args:
            url (str): picture url
            path (str): destination path
            name (str): filename
        '''
        if not os.path.exists(path): os.makedirs(path)
        full_path = os.path.join(path, name)
        if not os.path.exists(full_path):
            self.__download(url, full_path)              


    def __getPictureName(self,url):
        '''
        Get image name from url (ie my_movie_poster.jpg)      
        
        Args:
            url (str): full url of image                        
        Returns:
            picname (str) representing image name from file
        '''
        return re.split('\/+', url)[-1]
         
        
    def __download(self,url,mypath):
        '''
        Downloads images to store locally       
        
        Args:
            url (str): url of image to download
            mypath (str): local path to save image to
        '''        
        #print 'Attempting to download image from url: %s ' % url
        #print 'Saving to destination: %s ' % mypath
        if url.startswith('http://'):
            try:
                data = self.opener.open(url).read()
                fh = open(mypath, 'wb')
                fh.write(data)  
                fh.close()
            except Exception, e:
                #print 'Image download failed: %s ' % e
                pass
        else:
            if url is not None:
                #print 'Not a valid url: %s ' % url
                pass


    def __validIMDBId(self, imdb_id):
        '''
        Check and return a valid IMDB ID    
        
        Args:
            imdb_id (str): IMDB ID
        Returns:
            imdb_id (str) if valid with leading tt, else None
        '''      
        # add the tt if not found. integer aware.       
        if not imdb_id.startswith('tt'): imdb_id = 'tt%s' % imdb_id
        if re.search('tt[0-9]{7}', imdb_id): return imdb_id
        else: return None


    def __removeNoneValues(self, meta):
        ''' Ensure we are not sending back any None values, XBMC doesn't like them '''
        for item in meta:
            if meta[item] is None:
                meta[item] = ''          
        return meta


    def getMeta(self, type, name, imdb_id='', tmdb_id='', year=''):
        '''
        Main method to get meta data for movie or tvshow. Will lookup by name/year 
        if no IMDB ID supplied.       
        
        Args:
            type (str): 'movie' or 'tvshow'
            name (str): full name of movie/tvshow you are searching            
        Kwargs:
            imdb_id (str): IMDB ID        
            tmdb_id (str): TMDB ID
            year (str): 4 digit year of video, recommended to include the year whenever possible
                        to maximize correct search results.
                        
        Returns:
            DICT of meta data or None if cannot be found.
        '''
        try:
            name = name.encode('utf-8','replace')
        except:
            pass
        #print '---------------------------------------------------------------------------------------'
        #print 'Attempting to retreive meta data for %s: %s %s %s %s' % (type, name, year, imdb_id, tmdb_id)
        
        if imdb_id:
            imdb_id = self.__validIMDBId(imdb_id)

        if imdb_id:
            meta = self.__lookupById(type, imdb_id=imdb_id)
        elif tmdb_id:
            meta = self.__lookupById(type, tmdb_id=tmdb_id)
        else:
            meta = self.__lookupByName(type, name, year)

        if not meta:
            if type==self.typeMovie:       meta = self.__getTMDBMeta(imdb_id, tmdb_id, name, year)
            elif type==self.typeTVShow:    meta = self.__getTVDBMeta(imdb_id, name, year)
            self.__saveVideoMeta(meta, name, type)

        #We want to send back the name that was passed in   
        meta['title'] = name
              
        #Change cast back into a tuple
        if meta['cast']: meta['cast'] = eval(meta['cast'])
            
        #Return a trailer link that will play via youtube addon
        try:
            meta['trailer'] = 'plugin://plugin.video.youtube/?action=play_video&videoid=%s' % re.match('^[^v]+v=(.{3,11}).*', meta['trailer_url']).group(1)
        except:
            meta['trailer'] = ''
        
        #Ensure we are not sending back any None values, XBMC doesn't like them
        meta = self.__removeNoneValues(meta)
        
        #Add TVShowTitle infolabel
        if type==self.typeTVShow: meta['TVShowTitle'] = meta['title']
        
        #if cache row says there are pre-packed images then either use them or create them
        if meta['imgs_prepacked'] == 'true':
                #define the image paths               
                if type == self.typeMovie:
                    root_covers = self.mvcovers
                    root_backdrops = self.mvbackdrops
                elif type == self.typeTVShow:
                    root_covers = self.tvcovers
                    root_backdrops = self.tvbackdrops
                    root_banners = self.tvbanners
                
                if meta['cover_url']:
                    cover_name = self.__getPictureName(meta['cover_url'])
                    cover_path = os.path.join(root_covers, cover_name[0].lower())
                    if self.classmode == 'true':        self.__downloadImage(meta['cover_url'], cover_path, cover_name)
                    if os.path.exists(cover_name):      meta['cover_url'] = os.path.join(cover_path, cover_name)
                
                if meta['backdrop_url']:
                    backdrop_name = self.__getPictureName(meta['backdrop_url'])
                    backdrop_path=os.path.join(root_backdrops, backdrop_name[0].lower())
                    if self.classmode == 'true':        self.__downloadImage(meta['backdrop_url'], backdrop_path, backdrop_name)
                    if os.path.exists(backdrop_name):   meta['backdrop_url'] = os.path.join(backdrop_path, backdrop_name)

                if meta.has_key('banner_url'):
                    if meta['banner_url']:
                        banner_name = self.__getPictureName(meta['banner_url'])
                        banner_path=os.path.join(root_banners, banner_name[0].lower())
                        if self.classmode == 'true':        self.__downloadImage(meta['banner_url'], banner_path, banner_name)
                        if os.path.exists(banner_name):     meta['banner_url'] = os.path.join(banner_path, banner_name)        

        #print 'Returned Meta:', meta
        return meta  


    def updateMeta(self, type, name, imdb_id, tmdb_id='', new_imdb_id='', new_tmdb_id='', year=''):
        '''
        Updates and returns meta data for given movie/tvshow, mainly to be used with refreshing individual movies.
        
        Searches local cache DB for record, delete if found, calls getMeta() to grab new data

        name, imdb_id, tmdb_id should be what is currently in the DB in order to find current record
        
        new_imdb_id, new_tmdb_id should be what you would like to update the existing DB record to, which you should have already found
        
        Args:
            name (int): full name of movie you are searching            
            imdb_id (str): IMDB ID of CURRENT entry
        Kwargs:
            tmdb_id (str): TMDB ID of CURRENT entry
            new_imdb_id (str): NEW IMDB_ID to search with
            new_tmdb_id (str): NEW TMDB ID to search with
            year (str): 4 digit year of video, recommended to include the year whenever possible
                        to maximize correct search results.
                        
        Returns:
            DICT of meta data or None if cannot be found.
        '''
        try:
            name = name.encode('utf-8','replace')
        except:
            pass
        #print '---------------------------------------------------------------------------------------'
        #print 'Updating meta data: %s Old: %s %s New: %s %s Year: %s' % (name, imdb_id, tmdb_id, new_imdb_id, new_tmdb_id, year)
        
        if imdb_id:
            imdb_id = self.__validIMDBId(imdb_id)        
        
        if imdb_id:
            meta = self.__lookupById(type, imdb_id=imdb_id)
        elif tmdb_id:
            meta = self.__lookupById(type, tmdb_id=tmdb_id)
        else:
            meta = self.__lookupByName(type, name, year)
        
        if meta:
            self.__deleteVideoMeta(type, imdb_id, tmdb_id, name, year)
        else:
            #print 'No match found in cache db'
            pass
        
        if not new_imdb_id:
            new_imdb_id = imdb_id
        elif not new_tmdb_id:
            new_tmdb_id = tmdb_id
            
        return self.getMeta(type, name, new_imdb_id, new_tmdb_id, year)


    def __lookupById(self, type, imdb_id='', tmdb_id=''):
        '''
        Lookup in SQL DB for video meta data by IMDB ID
        
        Args:
            imdb_id (str): IMDB ID
            type (str): 'movie' or 'tvshow'
        Kwargs:
            imdb_id (str): IDMB ID
            tmdb_id (str): TMDB ID                        
        Returns:
            DICT of matched meta data or None if no match.
        '''   
        #print 'Looking up in local cache by id for: %s %s %s' % (type, imdb_id, tmdb_id)
        matchedrow = None
        try:
            
            if type == self.typeMovie:
                if imdb_id:
                    self.dbcur.execute("SELECT * FROM movie_meta WHERE imdb_id = ? LIMIT 1",(imdb_id,))
                else:
                    self.dbcur.execute("SELECT * FROM movie_meta WHERE tmdb_id = ? LIMIT 1",(tmdb_id,))
            elif type == self.typeTVShow:
                self.dbcur.execute("SELECT * FROM tvshow_meta  WHERE imdb_id = ? LIMIT 1",(imdb_id,))
            matchedrow = self.dbcur.fetchone()
            
            if matchedrow:
                #print 'Found meta information by id in cache table: ', dict(matchedrow)
                return dict(matchedrow)
            else:
                #print 'No match in local DB'
                return None
        except Exception, e:
            #print '************* Error selecting from cache db: %s' % e
            return None    

    def __lookupByName(self, type, name, year=''):
        '''
        Lookup in SQL DB for video meta data by name and year
        
        Args:
            type (str): 'movie' or 'tvshow'
            name (str): full name of movie/tvshow you are searching
        Kwargs:
            year (str): 4 digit year of video, recommended to include the year whenever possible
                        to maximize correct search results.
                        
        Returns:
            DICT of matched meta data or None if no match.
        '''        
        matchedrow = None
        try:
            #print 'Looking up in local cache by name for: %s %s %s' % (type, name, year)
            
            name =  self.__cleanString(name.lower())
            
            
            if type == self.typeMovie:
                if year:
                    self.dbcur.execute("SELECT * FROM movie_meta WHERE title = ?  AND year = ? LIMIT 1",(name,year))
                else:
                    self.dbcur.execute("SELECT * FROM movie_meta WHERE title = ? LIMIT 1",(name,))
            elif type == self.typeTVShow:
                self.dbcur.execute("SELECT * FROM tvshow_meta WHERE title = ? LIMIT 1",(name,))
            matchedrow = self.dbcur.fetchone()
            
            if matchedrow:
                #print 'Found meta information by name in cache table: ', dict(matchedrow)
                return dict(matchedrow)
            else:
                #print 'No match in local DB'            
                return None
        except Exception, e:
            #print '************* Error selecting from cache db: %s' % e
            pass
            



    def __saveVideoMeta(self, meta, name, type):
        '''
        Saves meta data to SQL table given type
        
        Args:
            meta (dict): meta data of video to be added to database
            type (str): 'movie' or 'tvshow'                       
        '''
        if type == self.typeMovie:
            table='movie_meta'
        elif type == self.typeTVShow:
            table='tvshow_meta'
        
        #strip title
        meta['title'] =  self.__cleanString(name.lower())
        
        matchedrow = None
        try:
            #print 'Checking if entry already exists in cache table: %s' % table

            #Select on either IMDB ID or name + premiered
            
            if meta['imdb_id']:
                self.dbcur.execute("SELECT * FROM %s WHERE imdb_id = '%s'" % table,(meta['imdb_id'],))
            else:           
                if meta.has_key('year') and meta['year']:
                    self.dbcur.execute("SELECT * FROM %s WHERE title = ? AND year = ?" % table, (meta['title'], meta['year']))
                else:
                    self.dbcur.execute("SELECT * FROM %s WHERE title = ?" % table, (meta['title'],))
            matchedrow = self.dbcur.fetchone()
            
        except Exception, e:
            #print '************* Error attempting to select from table: %s with error: %s' % (table, e)
            pass
            
        if matchedrow:
            try:
                #print 'Matched Row found, deleting table entry'
                self.dbcur.execute("DELETE FROM %s WHERE imdb_id = ?" % table, (meta['imdb_id'],))
            except Exception, e:
                #print '************* Error attempting to delete from cache table: %s ' % e
                #print 'Meta data:', meta               
                pass
        
        if meta.has_key('cast'):
            meta['cast'] = str(meta['cast'])
        
        #print 'Saving cache information: ', meta         
        try:
            
            if type == self.typeMovie:
                self.dbcur.execute("INSERT INTO %s VALUES (:imdb_id, :tmdb_id, :title, :year, :director, :writer, :tagline, :cast, :rating, :votes, :duration, :plot, :mpaa, :premiered, :genre, :studio, :thumb_url, :cover_url, :trailer_url, :backdrop_url, :imgs_prepacked)" % table, meta)
            elif type == self.typeTVShow:
                self.dbcur.execute("INSERT INTO %s VALUES (:imdb_id, :tvdb_id, :title, :cast, :rating, :duration, :plot, :mpaa, :premiered, :genre, :studio, :status, :banner_url, :cover_url, :trailer_url, :backdrop_url, :imgs_prepacked)" % table, meta)
            
        except Exception, e:
            #print '************* Error attempting to insert into cache table: %s ' % e
            #print 'Meta data:', meta
            pass
    
    
    def __deleteVideoMeta(self, type, imdb_id, tmdb_id, name, year):
        '''
        Delete meta data from SQL table
        
        Args:
            type (str): 'movie' or 'tvshow'
            imdb_id (str): IMDB ID
            tmdb_id (str): TMDB ID   
            name (str): Full movie name
            year (int): Movie year
                        
        '''         
        
        if type == self.typeMovie:
            table = 'movie_meta'
        elif type == self.typeTVShow:
            table = 'tvshow_meta'
        
        try:
            #print 'Deleting table entry: %s %s %s %s ' % (imdb_id, tmdb_id, name, year)
            
            if imdb_id:
                self.dbcur.execute("DELETE FROM %s WHERE imdb_id = ?" % table,(imdb_id,))
            elif tmdb_id:
                self.dbcur.execute("DELETE FROM %s WHERE tmdb_id = ?" % table,(tmdb_id,))
            else:
                name =  self.__cleanString(name.lower())
                if year:
                    self.dbcur.execute("DELETE FROM %s WHERE title = ? AND year = ?" % table, (name,year))
                else:
                    self.dbcur.execute("DELETE FROM %s WHERE title = ?" % table, (name,))
            
        except Exception, e:
            #print '************* Error attempting to delete from cache table: %s ' % e          
            pass    
        

    def __getTMDBMeta(self, imdb_id, tmdb_id, name, year=''):
        '''
        Requests meta data from TMDB and creates proper dict to send back
        
        Args:
            imdb_id (str): IMDB ID
            name (str): full name of movie you are searching
        Kwargs:
            year (str): 4 digit year of movie, when imdb_id is not available it is recommended
                        to include the year whenever possible to maximize correct search results.
                        
        Returns:
            DICT. It must also return an empty dict when
            no movie meta info was found from tmdb because we should cache
            these "None found" entries otherwise we hit tmdb alot.
        '''        
        
        tmdb = TMDB()        
        meta = tmdb.tmdb_lookup(name,imdb_id,tmdb_id, year)       
        
        if meta is None:
            # create an empty dict so below will at least populate empty data for the db insert.
            meta = {}
 
        return self.__formatTMDBMeat(meta, imdb_id, name, year)
    

    def __formatTMDBMeat(self, md, imdb_id, name, year):
        '''
        Copy tmdb to our own for conformity and eliminate KeyError. Set default for values not returned
        
        Args:
            imdb_id (str): IMDB ID
            name (str): full name of movie you are searching
        Kwargs:
            year (str): 4 digit year of movie, when imdb_id is not available it is recommended
                        to include the year whenever possible to maximize correct search results.
                        
        Returns:
            DICT. It must also return an empty dict when
            no movie meta info was found from tvdb because we should cache
            these "None found" entries otherwise we hit tvdb alot.
        '''      
        
        #Intialize movie_meta dictionary    
        meta = self.__initMovieMeta(imdb_id, md.get('id', ''), name, year)
        
        meta['imdb_id'] = md.get('imdb_id', imdb_id)
        meta['title'] = md.get('name', name)      
        meta['tagline'] = md.get('tagline', '')
        meta['rating'] = float(md.get('rating', 0))
        meta['votes'] = str(md.get('votes', ''))
        meta['duration'] = str(md.get('runtime', 0))
        meta['plot'] = md.get('overview', '')
        meta['mpaa'] = md.get('certification', '')       
        meta['premiered'] = md.get('released', '')
        
        #Do whatever we can to set a year, if we don't have one lets try to strip it from premiered
        if not year and meta['premiered']:
            meta['year'] = int(self.__convertDate(meta['premiered'], '%Y-%m-%d', '%Y'))
                   
        meta['trailer_url'] = md.get('trailer', '')
        meta['genre'] = md.get('genre', '')
        
        #Get cast, director, writers
        cast_list = []
        cast_list = md.get('cast','')
        if cast_list:
            for cast in cast_list:
                job=cast.get('job','')
                if job == 'Actor':
                    meta['cast'].append((cast.get('name',''),cast.get('character','') ))
                elif job == 'Director':
                    meta['director'] = cast.get('name','')
                elif job == 'Screenplay':
                    if meta['writer']:
                        meta['writer'] = meta['writer'] + ' / ' + cast.get('name','')
                    else:
                        meta['writer'] = cast.get('name','')
                    
        genre_list = []
        genre_list = md.get('genres', '')
        for genre in genre_list:
            if meta['genre'] == '':
                meta['genre'] = genre.get('name','')
            else:
                meta['genre'] = meta['genre'] + ' / ' + genre.get('name','')
        
        if md.has_key('tvdb_studios'):
            meta['studio'] = md.get('tvdb_studios', '')
        try:
            meta['studio'] = (md.get('studios', '')[0])['name']
        except:
            try:
                meta['studio'] = (md.get('studios', '')[1])['name']
            except:
                try:
                    meta['studio'] = (md.get('studios', '')[2])['name']
                except:
                    try:    
                        meta['studio'] = (md.get('studios', '')[3])['name']
                    except:
                        #print 'Studios failed: %s ' % md.get('studios', '')
                        pass
        
        meta['cover_url'] = md.get('cover_url', '')
        if md.has_key('posters'):
            # find first thumb poster url
            for poster in md['posters']:
                if poster['image']['size'] == 'thumb':
                    meta['thumb_url'] = poster['image']['url']
                    break
            # find first cover poster url
            for poster in md['posters']:
                if poster['image']['size'] == 'cover':
                    meta['cover_url'] = poster['image']['url']
                    break

        if md.has_key('backdrops'):
            # find first original backdrop url
            for backdrop in md['backdrops']:
                if backdrop['image']['size'] == 'original':
                    meta['backdrop_url'] = backdrop['image']['url']
                    break

        return meta
        
        
    def __getTVDBMeta(self, imdb_id, name, year=''):
        '''
        Requests meta data from TVDB and creates proper dict to send back
        
        Args:
            imdb_id (str): IMDB ID
            name (str): full name of movie you are searching
        Kwargs:
            year (str): 4 digit year of movie, when imdb_id is not available it is recommended
                        to include the year whenever possible to maximize correct search results.
                        
        Returns:
            DICT. It must also return an empty dict when
            no movie meta info was found from tvdb because we should cache
            these "None found" entries otherwise we hit tvdb alot.
        '''      
        #print 'Starting TVDB Lookup'
        tvdb = TheTVDB()
        tvdb_id = ''
        
        if imdb_id:
            tvdb_id = tvdb.get_show_by_imdb(imdb_id)
            
        #Intialize tvshow meta dictionary
        meta = self.__initTVShowMeta(imdb_id, tvdb_id, name)

        # if not found by imdb, try by name
        if tvdb_id == '':
            try:
                #If year is passed in, add it to the name for better TVDB search results
                if year:
                    name = name + ' ' + year
                show_list=tvdb.get_matching_shows(name)
            except Exception, e:
                #print '************* Error retreiving from thetvdb.com: %s ' % e     
                show_list = []
                pass
            #print 'Found TV Show List: ', show_list
            tvdb_id=''
            prob_id=''
            for show in show_list:
                (junk1, junk2, junk3) = show
                #if we match imdb_id or full name (with year) then we know for sure it is the right show
                if junk3==imdb_id or self.__stringCompare(self.__cleanString(junk2),self.__cleanString(name)):
                    tvdb_id=self.__cleanString(junk1)
                    if not imdb_id:
                        imdb_id=self.__cleanString(junk3)
                    break
                #if we match just the cleaned name (without year) keep the tvdb_id
                elif self.__stringCompare(self.__cleanString(junk2),self.__cleanString(name)):
                    prob_id = junk1
                    if not imdb_id:
                        imdb_id = self.__cleanString(junk3)
            if tvdb_id == '' and prob_id != '':
                tvdb_id = self.__cleanString(prob_id)

        if tvdb_id:
            #print 'Show *** ' + name + ' *** found in TVdb. Getting details...'
            show = tvdb.get_show(tvdb_id)
            if show is not None:
                meta['imdb_id'] = imdb_id
                meta['tvdb_id'] = tvdb_id
                meta['title'] = name
                meta['TVShowTitle'] = name
                if str(show.rating) != '' and show.rating != None:
                    meta['rating'] = float(show.rating)
                meta['duration'] = show.runtime
                meta['plot'] = show.overview
                meta['mpaa'] = show.content_rating
                meta['premiered'] = str(show.first_aired)
                if show.genre != '':
                    temp = show.genre.replace("|",",")
                    temp = temp[1:(len(temp)-1)]
                    meta['genre'] = temp
                meta['studio'] = show.network
                meta['status'] = show.status
                if show.actors:
                    for actor in show.actors:
                        meta['cast'].append(actor)
                meta['banner_url'] = show.banner_url
                meta['imgs_prepacked'] = self.classmode
                meta['cover_url'] = show.poster_url
                meta['backdrop_url'] = show.fanart_url

                if meta['plot'] == 'None' or meta['plot'] == '' or meta['plot'] == 'TBD' or meta['plot'] == 'No overview found.' or meta['rating'] == 0 or meta['duration'] == 0 or meta['cover_url'] == '':
                    #print ' Some info missing in TVdb for TVshow *** '+ name + ' ***. Will search imdb for more'                    
                    tmdb = TMDB()
                    imdb_meta = tmdb.search_imdb(name, imdb_id)
                    if imdb_meta:
                        imdb_meta = tmdb.update_imdb_meta(meta, imdb_meta)
                        if imdb_meta.has_key('overview'):
                            meta['plot'] = imdb_meta['overview']
                        if imdb_meta.has_key('rating'):
                            meta['rating'] = imdb_meta['rating']
                        if imdb_meta.has_key('runtime'):
                            meta['duration'] = imdb_meta['runtime']
                        if imdb_meta.has_key('cast'):
                            meta['cast'] = imdb_meta['cast']
                        if imdb_meta.has_key('cover_url'):
                            meta['cover_url'] = imdb_meta['cover_url']

                return meta
            else:
                tmdb = TMDB()
                imdb_meta = tmdb.search_imdb(name, imdb_id)
                if imdb_meta:
                    meta = tmdb.update_imdb_meta(meta, imdb_meta)
                return meta    
        else:
            return meta

    
    def searchMovies(self, name):
        '''
        Requests meta data from TMDB for any movie matching name
        
        Args:
            name (str): full name of movie you are searching
                        
        Returns:
            Arry of dictionaries with trimmed down meta data, only returned data that is required:
            - IMDB ID
            - TMDB ID
            - Name
            - Year
        ''' 
        try:
            name = name.encode('utf-8','replace')
        except:
            pass
        #print '---------------------------------------------------------------------------------------'
        #print 'Meta data refresh - searching for movie: %s' % name
        tmdb = TMDB()
        movie_list = []
        meta = tmdb.tmdb_search(name)
        if meta:
            for movie in meta:
                if movie['released']:
                    year = self.__convertDate(movie['released'], '%Y-%m-%d', '%Y')
                else:
                    year = None
                movie_list.append({'title': movie['name'], 'imdb_id': movie['imdb_id'], 'tmdb_id': movie['id'], 'year': year})
        else:
            #print 'No results found'
            return None
 
        #print 'Returning results: %s' % movie_list
        return movie_list

            
    def getEpisodeMeta(self, name, imdb_id, season, episode):
        '''
        Requests meta data from TVDB for TV episodes, searches local cache db first.
        
        Args:
            name (str): full name of tvshow you are searching
            imdb_id (str): IMDB ID
            season (int): tv show season number, number only no other characters
            episode (int): tv show episode number, number only no other characters
                        
        Returns:
            DICT. It must also return an empty dict when
            no meta info was found in order to save these.
        '''  
        try:
            name = name.encode('utf-8','replace')
        except:
            pass
        #print '---------------------------------------------------------------------------------------'
        #print 'Attempting to retreive episode meta data for: %s %s %s' % (imdb_id, season, episode)
        try:
            dateSearch = False
            searchTVDB = True
            
            if imdb_id:
                imdb_id = self.__validIMDBId(imdb_id)
                    
            #Find tvdb_id for the TVshow
            tvdb_id = self.__getTVDBId(name, imdb_id)
            
            #Check if it exists in local cache first
            meta = self.__lookupEpisode(imdb_id, tvdb_id, season, episode)
            
            #If not found lets scrape online sources
            if meta is None:
                
                if tvdb_id == '' or tvdb_id is None:
                    #print "Could not find TVshow with imdb: %s " % imdb_id
                    
                    meta = {}
                    meta['imdb_id']=imdb_id
                    meta['tvdb_id']=''
                    meta['episode_id'] = ''                
                    meta['season']=season
                    meta['episode']=episode
                    meta['title']= name
                    meta['plot'] = ''
                    meta['director'] = ''
                    meta['writer'] = ''
                    meta['rating'] = 0
                    meta['premiered'] = ''
                    meta['poster'] = ''
                    meta['cover_url']=meta['poster']
                    meta['trailer_url']=''
                    meta['premiered']=meta['premiered']
                    meta = self.__getTVExtra(meta)
                    meta['backdrop_url'] = ''
                        
                    self.__saveEpisodeMeta(meta)
                    
                    return meta
                
                if searchTVDB:
                    meta = self.__getTVDBEpisodeMeta(tvdb_id, season, episode, dateSearch)
                    if meta is None:
                        meta = {}
                        meta['title']= name
                        meta['director'] = ''
                        meta['writer'] = ''                    
                        meta['episode_id'] = ''
                        meta['plot'] = ''
                        meta['rating'] = 0
                        meta['premiered'] = ''
                        meta['poster'] = ''
                        meta['season'] = 0
                        meta['episode'] = 0
                        meta['backdrop_url'] = ''                    
                else:
                    meta = {}
                    meta['title']= name
                    meta['episode_id'] = ''
                    meta['plot'] = ''
                    meta['director'] = ''
                    meta['writer'] = ''                
                    meta['rating'] = 0
                    meta['premiered'] = ''
                    meta['poster'] = ''
                    meta['season'] = 0
                    meta['episode'] = 0
                    meta['backdrop_url'] = ''                
                    
                #if meta is not None:
                if not meta['title']:
                    meta['title']= name
                meta['imdb_id']=imdb_id
                meta['tvdb_id']=tvdb_id
                meta['season']=int(season)
                meta['episode']=int(episode)
                meta['cover_url']=meta['poster']
                meta['trailer_url']=''
                meta['premiered']=meta['premiered']
                meta = self.__getTVExtra(meta)
                
                self.__saveEpisodeMeta(meta)
                
                meta['backdrop_url'] = self.__getTVShowBackDrops(imdb_id, tvdb_id)
            
            else:
                #print 'Episode found on db, meta='+str(meta)
                pass
    
            #Ensure we are not sending back any None values, XBMC doesn't like them
            meta = self.__removeNoneValues(meta)
            
            if meta.get("title",None) != None and meta.get("episode",None) != None:
                if not "Episode %02d:" % meta.get("episode") in meta.get("title"):
                    meta["title"] = "Episode %02d: %s" % (meta.get("episode"),meta.get("title"))
        except:
            pass
        return meta
  
    
    def __getTVExtra(self, meta):
        '''
        When requesting episode information, not all data may be returned
        Fill in extra missing meta information from tvshow_meta table which should
        have already been populated.
        
        Args:
            meta (dict): current meta dict
                        
        Returns:
            DICT containing the extra values
        '''
        #print 'Retrieving extra TV Show information from tvshow_meta'        
        try:
            
            if meta['imdb_id']:
                self.dbcur.execute("SELECT * FROM tvshow_meta WHERE imdb_id = ?",(meta['imdb_id'],))
            elif meta['tvdb_id']:
                self.dbcur.execute("SELECT * FROM tvshow_meta WHERE tvdb_id = ?",(meta['tvdb_id'],))
            else:
                self.dbcur.execute("SELECT * FROM tvshow_meta WHERE title = ?",(self.__cleanString(meta['title'].lower()),))
            matchedrow = self.dbcur.fetchone()
            
            if matchedrow:
                match = dict(matchedrow)
                meta['genre'] = match['genre']
                meta['duration'] = match['duration']
                meta['studio'] = match['studio']
                meta['mpaa'] = match['mpaa']
            else:
                meta['genre'] = ''
                meta['duration'] = '0'
                meta['studio'] = ''
                meta['mpaa'] = ''
        except Exception, e:
            #print '************* Error attempting to select from tvshow_meta table: %s ' % e          
            pass
        return meta


    def __getTVDBId(self, name, imdb_id):
        '''
        Retrieves TVID for a tv show that has already been scraped and saved in cache db.
        
        Used when scraping for season and episode data
        
        Args:
            name (str): full name of tvshow you are searching            
            imdb_id (str): IMDB ID
                        
        Returns:
            (str) imdb_id 
        '''      
        
        #clean tvshow name of any extras       
        name =  self.__cleanString(name.lower())
        
        #print 'Retrieving TVDB ID'
        try:
            
            if imdb_id:
                self.dbcur.execute("SELECT tvdb_id FROM tvshow_meta WHERE imdb_id = ?" , (imdb_id,))
            elif name:
                self.dbcur.execute("SELECT tvdb_id FROM tvshow_meta WHERE title = ?" ,(name,))
            matchedrow = self.dbcur.fetchone()
            
        except Exception, e:
            #print '************* Error attempting to select from tvshow_meta table: %s ' % e          
            pass
                        
        if matchedrow:
            return dict(matchedrow)['tvdb_id']
        else:
            return None


    def updateEpisodeMeta(self, name, imdb_id, season, episode, tvdb_id='', new_imdb_id='', new_tvdb_id=''):
        '''
        Updates and returns meta data for given episode, 
        mainly to be used with refreshing individual tv show episodes.
        
        Searches local cache DB for record, delete if found, calls getEpisodeMeta() to grab new data
               
        
        Args:
            name (int): full name of movie you are searching
            imdb_id (str): IMDB ID
            season (int): season number
            episode (int): episode number
        Kwargs:
            tvdb_id (str): TVDB ID
                        
        Returns:
            DICT of meta data or None if cannot be found.
        '''
        try:
            name = name.encode('utf-8','replace')
        except:
            pass
        #print '---------------------------------------------------------------------------------------'
        #print 'Updating episode meta data: %s IMDB: %s SEASON: %s EPISODE: %s TVDB ID: %s NEW IMDB ID: %s NEW TVDB ID: %s' % (name, imdb_id, season, episode, tvdb_id, new_imdb_id, new_tvdb_id)

      
        if imdb_id:
            imdb_id = self.__validIMDBId(imdb_id)
        else:
            imdb_id = ''

        #Find tvdb_id for the TVshow
        tvdb_id = self.__getTVDBId(name, imdb_id)
        
        #Lookup in cache table for existing entry
        meta = self.__lookupEpisode(imdb_id, tvdb_id, season, episode)
        
        #We found an entry in the DB, so lets delete it
        if meta:
            self.__deleteEpisodeMeta(imdb_id, tvdb_id, name, season, episode)
        else:
            #print 'No match found in cache db'
            pass
       
        if not new_imdb_id:
            new_imdb_id = imdb_id
        elif not new_tvdb_id:
            new_tvdb_id = tvdb_id
            
        return self.getEpisodeMeta(name, imdb_id, season, episode)


    def __lookupEpisode(self, imdb_id, tvdb_id, season, episode):
        '''
        Lookup in local cache db for episode data
        
        Args:
            imdb_id (str): IMDB ID
            tvdb_id (str): TheTVDB ID
            season (str): tv show season number, number only no other characters
            episode (str): tv show episode number, number only no other characters
                        
        Returns:
            DICT. Returns results found or None.
        ''' 
        #print 'Looking up episode data in cache db, imdb id: %s season: %s episode: %s' % (imdb_id, season, episode) 
        try:
            self.dbcur.execute(""" SELECT 
                               episode_meta.title as title, 
                               episode_meta.plot as plot, 
                               episode_meta.director as director, 
                               episode_meta.writer as writer, 
                               tvshow_meta.genre as genre, 
                               tvshow_meta.duration as duration, 
                               episode_meta.premiered as premiered, 
                               tvshow_meta.studio as studio, 
                               tvshow_meta.mpaa as mpaa, 
                               episode_meta.imdb_id as imdb_id, 
                               episode_meta.rating as rating, 
                               "" as trailer_url, 
                               episode_meta.season as season, 
                               episode_meta.episode as episode, 
                               tvshow_meta.backdrop_url as backdrop_url,                                
                               episode_meta.poster as cover_url  
                               FROM episode_meta, tvshow_meta WHERE 
                               episode_meta.imdb_id = tvshow_meta.imdb_id AND 
                               episode_meta.tvdb_id = tvshow_meta.tvdb_id AND 
                               episode_meta.imdb_id = ? AND episode_meta.tvdb_id = ? AND season = ? AND episode_meta.episode = ?
                            """,(imdb_id, tvdb_id, season, episode))
            matchedrow = self.dbcur.fetchone()
            if matchedrow:
                #print 'Found episode meta information in cache table: ', dict(matchedrow)
                return dict(matchedrow)
            else:
                return None
        except Exception, e:
            #print '************* Error attempting to select from Episode table: %s ' % e
            pass
        

    def __deleteEpisodeMeta(self, imdb_id, tvdb_id, name, season, episode):
        '''
        Delete meta data from SQL table
        
        Args:
            imdb_id (str): IMDB ID
            tvdb_id (str): TVDB ID
            name (str): Episode title
            season (int): Season #
            episode(int): Episode #
        '''
        #print 'Deleting table entry: IMDB: %s TVDB: %s Title: %s Season: %s Episode: %s ' % (imdb_id, tvdb_id, name, season, episode)            
        try:
            self.dbcur.execute("DELETE FROM episode_meta WHERE imdb_id = ? AND tvdb_id = ? and season = ? and episode = ?",(imdb_id, tvdb_id, season, episode))
        except Exception, e:
            #print '************* Error attempting to delete from episode cache table: %s ' % e          
            pass


    def __getTVDBEpisodeMeta(self, tvdb_id, season, episode, dateSearch=False):
        '''
        Initiates lookup for episode data on TVDB
        
        Args:
            tvdb_id (str): TVDB id
            season (str): tv show season number, number only no other characters
            episode (str): tv show episode number, number only no other characters
            dateSearch (bool): search based on a date range
                        
        Returns:
            DICT. Data found from lookup
        '''      
        #get metadata text using themoviedb api
        meta = self.__TVDBLookup(tvdb_id,season,episode, dateSearch)      
        return meta


    def __TVDBLookup(self, tvdb_id, season_num, episode_num, dateSearch):
        #TvDB Lookup for episodes
        
        meta = {}
        tvdb = TheTVDB()
        if dateSearch:
            aired=self.__getDate(season_num, episode_num)
            episode = tvdb.get_episode_by_airdate(tvdb_id, aired)
            
            #We do this because the airdate method returns just a part of the overview unfortunately
            if episode is not None:
                ep_id = episode.id
                if ep_id is not None:
                    episode = tvdb.get_episode(ep_id)
        else:
            episode = tvdb.get_episode_by_season_ep(tvdb_id, season_num, episode_num)
            
        if episode is None:
            return None
        
        meta['episode_id'] = episode.id
        meta['plot'] = self.__check(episode.overview)
        if episode.guest_stars:
            guest_stars = episode.guest_stars.replace('|', ', ')
            meta['plot'] = meta['plot'] + '\n\nGuest Starring: ' + guest_stars
        meta['rating'] = float(self.__check(episode.rating,0))
        meta['premiered'] = self.__check(episode.first_aired)
        meta['title'] = self.__check(episode.name)
        meta['poster'] = self.__check(episode.image)
        meta['director'] = self.__check(episode.director)
        meta['writer'] = self.__check(episode.writer)
        meta['season'] = int(self.__check(episode.season_number,0))
        meta['episode'] = int(self.__check(episode.episode_number,0))
              
        return meta


    def __check(self, value, ret=None):
        if value is None or value == '':
            if ret == None:
                return ''
            else:
                return ret
        else:
            return value
            
        
    def __saveEpisodeMeta(self, meta):
        '''
        Save episode data to local cache db.
        
        Args:
            meta (dict): episode data to be stored
                        
        ''' 
        try: 
            if meta['imdb_id']:
                self.dbcur.execute("DELETE FROM episode_meta WHERE imdb_id = ? AND season = ? AND episode = ?",(meta['imdb_id'], meta['season'], meta['episode']))
            elif meta['tvdb_id']:
                self.dbcur.execute("DELETE FROM episode_meta WHERE tvdb_id = ? AND season = ? AND episode = ?",(meta['tvdb_id'], meta['season'], meta['episode']))
            else:         
                self.dbcur.execute("DELETE FROM episode_meta WHERE title = ? AND season = ? AND episode = ?",(self.__cleanString(meta['title'].lower()), meta['season'], meta['episode']))
            #print 'Saving Episode Meta'
        except Exception, e:
            #print '************* Error attempting to delete from cache table: %s ' % e
            #print 'Meta data:', meta
            pass        
        
        #print 'Saving episode cache information: ', meta
        try:
            self.dbcur.execute("INSERT INTO episode_meta VALUES (:imdb_id, :tvdb_id, :episode_id, :season, :episode, :title, :director, :writer, :plot, :rating, :premiered, :poster)",meta)
        except Exception, e:
            #print '************* Error attempting to insert into cache table: %s ' % e
            #print 'Meta data:', meta
            pass


    def updateTrailer(self, type, imdb_id, trailer, tmdb_id=''):
        '''
        Change watched status on video
        
        Args:
            type (str): type of video to update, 'movie', 'tvshow' or 'episode'
            imdb_id (str): IMDB ID
            trailer (str): url of youtube video
        Kwargs:            
            tmdb_id (str): TMDB ID
                        
        '''      
        if type == 'movie':
            table='movie_meta'
        elif type == 'tvshow':
            table='tvshow_meta'
        
        if imdb_id:
            imdb_id = self.__validIMDBId(imdb_id)

        #print 'Updating trailer for type: %s, imdb id: %s, tmdb_id: %s, trailer: %s' % (type, imdb_id, tmdb_id, trailer)
        try:
            if imdb_id:
                self.dbcur.execute("UPDATE %s set trailer_url=? WHERE imdb_id = ?" % table, (trailer, imdb_id))
            elif tmdb_id:
                self.dbcur.execute("UPDATE %s set trailer_url=? WHERE tmdb_id = ?" % table, (trailer, tmdb_id))
        except Exception, e:
            #print '************* Error attempting to update table: %s ' % e
            pass

    def __findCover(self, season, images):
        '''
        Finds the url of the banner to be used as the cover 
        from a list of images for a given season
        
        Args:
            season (str): tv show season number, number only no other characters
            images (dict): all images related
                        
        Returns:
            (str) cover_url: url of the selected image
        '''         
        cover_url = ''
        
        for image in images:
            (banner_url, banner_type, banner_season) = image
            if banner_season == season and banner_type == 'season':
                cover_url = banner_url
                break
        
        return cover_url
    

    def getSeasons(self, name, imdb_id, seasons):
        '''
        Requests from TVDB a list of images for a given tvshow
        and list of seasons
        
        Args:
            imdb_id (str): IMDB ID
            seasons (str): a list of seasons, numbers only
                        
        Returns:
            (list) list of covers found for each season
        '''     
        if imdb_id:
            imdb_id = self.__validIMDBId(imdb_id)
            
        if type(seasons) == str:
            seasons = [seasons]
                
        coversList = []
        tvdb_id = self.__getTVDBId(name, imdb_id)
        images  = None
        for season in seasons:
            meta = self.__lookupSeason(imdb_id, tvdb_id, season)
            if meta is None:
                meta = {}
                if tvdb_id is None or tvdb_id == '':
                    meta['cover_url']=''
                elif images:
                    meta['cover_url']=self.__findCover(season, images )
                else:
                    if len(season) == 4:
                        meta['cover_url']=''
                    else:
                        images = self.__getSeasonPosters(tvdb_id, season)
                        meta['cover_url']=self.__findCover(season, images )
                        
                meta['season'] = int(season)
                meta['tvdb_id'] = tvdb_id
                meta['imdb_id'] = imdb_id
                meta['backdrop_url'] = self.__getTVShowBackDrops(imdb_id, tvdb_id)                    
                
                #Ensure we are not sending back any None values, XBMC doesn't like them
                meta = self.__removeNoneValues(meta)
                                
                self.__saveSeasonMeta(meta)
            
            coversList.append(meta)
                   
        return coversList


    def updateSeason(self, name, imdb_id, season):
        '''
        Update an individual season:
            - looks up and deletes existing entry
            - re-scans TVDB for season image
        
        Args:
            imdb_id (str): IMDB ID
            season (int): season number to be refreshed
                        
        Returns:
            (list) list of covers found for each season
        '''     

        #Find tvdb_id for the TVshow
        tvdb_id = self.__getTVDBId(name, imdb_id)
        #print '---------------------------------------------------------------------------------------'
        #print 'Updating season meta data: %s IMDB: %s TVDB ID: %s SEASON: %s' % (name, imdb_id, tvdb_id, season)

        if imdb_id:
            imdb_id = self.__validIMDBId(imdb_id)
        else:
            imdb_id = ''
       
        #Lookup in cache table for existing entry
        meta = self.__lookupSeason(imdb_id, tvdb_id, season)
        
        #We found an entry in the DB, so lets delete it
        if meta:
            self.__deleteSeasonMeta(imdb_id, tvdb_id, season)
        else:
            #print 'No match found in cache db'
            pass

        return self.getSeasons(name, imdb_id, season)


    def __getTVShowBackDrops(self, imdb_id, tvdb_id):
        '''
        Gets the backdrop_url from tvshow_meta to be included with season & episode meta
        
        Args:              
            imdb_id (str): IMDB ID
            tvdb_id (str): TVDB ID

        '''
        try:
            self.dbcur.execute("SELECT backdrop_url FROM tvshow_meta WHERE imdb_id = ? AND tvdb_id = ?",(imdb_id, tvdb_id))
            matchedrow = self.dbcur.fetchone()
            if matchedrow:
                return dict(matchedrow)['backdrop_url']
        except Exception, e:
            #print '************* Error attempting to select from tvshow_meta table: %s ' % e
            pass
        return ''
    
    
    def __getSeasonPosters(self, tvdb_id, season):
        tvdb = TheTVDB()
        images = tvdb.get_show_image_choices(tvdb_id)       
        return images
        

    def __lookupSeason(self, imdb_id, tvdb_id, season):
        '''
        Lookup data for a given season in the local cache DB.
        
        Args:
            imdb_id (str): IMDB ID
            tvdb_id (str): TVDB ID
            season (str): tv show season number, number only no other characters
                        
        Returns:
            (dict) meta data for a match
        '''      
        
        #print 'Looking up season data in cache db, imdb id: %s tvdb_id: %s season: %s' % (imdb_id, tvdb_id, season)
        try:
            if imdb_id:
                self.dbcur.execute("SELECT a.*, b.backdrop_url FROM season_meta a, tvshow_meta b WHERE a.imdb_id = ? AND season = ? and a.imdb_id=b.imdb_id and a.tvdb_id=b.tvdb_id",(imdb_id, season))
            elif tvdb_id:
                self.dbcur.execute("SELECT a.*, b.backdrop_url FROM season_meta a, tvshow_meta b WHERE a.tvdb_id = ? AND season = ?  and a.imdb_id=b.imdb_id and a.tvdb_id=b.tvdb_id",(tvdb_id, season))
                
            matchedrow = self.dbcur.fetchone()
            if matchedrow:
                #print 'Found season meta information in cache table: ', dict(matchedrow)
                return dict(matchedrow)
        except Exception, e:
            #print '************* Error attempting to select from season_meta table: %s ' % e
            pass 
        return None
    

    def __saveSeasonMeta(self, meta):
        '''
        Save data for a given season in local cache DB.
        
        Args:
            meta (dict): full meta data for season
        '''
        try:
            self.dbcur.execute("DELETE FROM season_meta WHERE imdb_id = ? AND season = ?",(meta['imdb_id'], meta['season'])) 
        except Exception, e:
            #print '************* Error attempting to delete from cache table: %s ' % e
            #print 'Meta data:', meta
            pass       
        #print 'Saving season cache information: ', meta
        try:
            self.dbcur.execute("INSERT INTO season_meta VALUES (:imdb_id, :tvdb_id, :season, :cover_url)",meta)
        except Exception, e:
            #print '************* Error attempting to insert into cache table: %s ' % e
            #print 'Meta data:', meta
            pass         


    def __deleteSeasonMeta(self, imdb_id, tvdb_id, season):
        '''
        Delete meta data from SQL table
        
        Args:
            imdb_id (str): IMDB ID
            tvdb_id (str): TVDB ID
            season (int): Season #
        '''
        #print 'Deleting table entry: IMDB: %s TVDB: %s Season: %s ' % (imdb_id, tvdb_id, season)             
        try:
            self.dbcur.execute("DELETE FROM season_meta WHERE imdb_id = ? AND tvdb_id = ? and season = ?",(imdb_id, tvdb_id, season))
        except Exception, e:
            #print '************* Error attempting to delete from season cache table: %s ' % e          
            pass
        
        
    